from firedrake import *
from firedrake.utils import RealType
import numpy as np

from FIAT.reference_element import UFCInterval
from FIAT.quadrature import GaussLegendreQuadratureLineRule
from finat.point_set import GaussLegendrePointSet
from finat.quadrature import QuadratureRule


roots = {1: [2.],
         2: [3. + 1.73205081 * 1j,
             3. - 1.73205081 * 1j],
         3: [4.64437071,
             3.67781465 + 3.50876192 * 1j,
             3.67781465 - 3.50876192 * 1j],
         4: [4.20757879 + 5.31483608 * 1j,
             4.20757879 - 5.31483608 * 1j,
             5.79242121 + 1.73446826 * 1j,
             5.79242121 - 1.73446826 * 1j]}


def compute_discrete_wave_numbers(gamma, h, p, reduced_rule):
    mesh = IntervalMesh(1, -h, 0.0, comm=COMM_SELF)
    V = FunctionSpace(mesh, "P", p)
    u = TrialFunction(V)
    v = TestFunction(V)
    a = (gamma**2 * inner(u, v) + inner(grad(u), grad(v))) * dx(scheme=reduced_rule)
    A = assemble(a)
    if p == 1:
        C = A.M.values
    else:
        A00 = A.M.values[0:(p - 1), 0:(p - 1)]  # cell DoFs
        A01 = A.M.values[0:(p - 1), (p - 1):(p + 1)]
        A10 = A.M.values[(p - 1):(p + 1), 0:(p - 1)]
        A11 = A.M.values[(p - 1):(p + 1), (p - 1):(p + 1)]
        A00_ = np.linalg.inv(A00)
        u0 = np.matmul(A00_, -A01)
        C = np.matmul(A10, u0) + A11
    xi = np.roots([C[0, 1], C[1, 1] + C[0, 0], C[1, 0]])
    if abs(xi[0]) > 1:
        xi_out, xi_in = xi[1], xi[0]
    else:
        xi_out, xi_in = xi[0], xi[1]
    return xi_out, xi_in


def compute_reflection(gamma, h, p, u, reduced_rule):
    xi_out, xi_in = compute_discrete_wave_numbers(gamma, h, p, reduced_rule)
    mat = np.array([[xi_out**(-1), xi_in**(-1)], [1., 1.]])
    vec = np.array([u.at(-h), u.at(0)])
    cc = np.matmul(np.linalg.inv(mat), vec)
    return abs(cc[1] / cc[0])


# Parameters
h = 1
x0 = -h
x1 = h
p = COMM_WORLD.rank + 1
gamma = Constant(0.0)
gamma_0 = Constant(exp(1j * pi * 1 / 3) * h)

mesh = IntervalMesh(round((x1 - x0) / h), x0, x1, comm=COMM_SELF)
V = FunctionSpace(mesh, "P", p)
u = TrialFunction(V)
v = TestFunction(V)
fiat_rule = GaussLegendreQuadratureLineRule(UFCInterval(), p)
point_set = GaussLegendrePointSet(fiat_rule.get_points())
reduced_rule = QuadratureRule(point_set, fiat_rule.get_weights())
# Use reduced rule everywhere.
x, = SpatialCoordinate(mesh)
DP0 = FunctionSpace(mesh, "DP", 0)
profile = Function(DP0).interpolate(conditional(x < 0, 1, gamma_0))
a = (gamma**2 / profile * inner(u, v) + profile * inner(grad(u), grad(v))) * dx(scheme=reduced_rule)
L = inner(Constant(0), v) * dx
bc1 = DirichletBC(V, 1.0, 1)
bc2 = DirichletBC(V, 0.0, 2)
u = Function(V)
prob = LinearVariationalProblem(a, L, u, [bc1, bc2])
solv = LinearVariationalSolver(prob, solver_parameters={"ksp_rtol": 1e-15})
#
nx_, ny = (20, 40)
nx = nx_ - 1
W = 8.
xmesh = np.linspace(W / nx_, W, nx)
ymesh = np.linspace(-W, W, ny)
xv, yv = np.meshgrid(xmesh, ymesh)
err = np.zeros(shape=(ny, nx), dtype=RealType)
for ix in range(nx):
    for iy in range(ny):
        print(f"p = {p}: {ix} / {nx}, {iy} / {ny}", flush=True)
        gamma.assign(xv[iy, ix] + 1j * yv[iy, ix])
        # Analytical solution:
        # u_exact = Function(V).interpolate(exp(-gamma * (x - x0)))
        solv.solve()
        error = compute_reflection(gamma, h, p, u, reduced_rule)
        err[iy, ix] = np.log10(error).real
with open(f"main_fig_example_1d_p{p}_rotated.txt", "w") as f:
    for ix in range(nx):
        for iy in range(ny):
            f.write(f"{xv[iy, ix]} {yv[iy, ix]} {err[iy, ix]}\n")
        f.write("\n")
#
for root in roots[p]:
    root_ = root * gamma_0.values() / h
    # Add perturbation.
    # Incoming/Outgoing waves are not defined right at root points
    # for the problem where reduced integration is used on the entire domain.
    gamma.assign(root_ * (1 + 1.e-6))
    solv.solve()
    error = compute_reflection(gamma, h, p, u, reduced_rule)
    print(f"p = {p}: root = {root_}, error = {np.log10(error).real}")
if False:
    from matplotlib import pyplot as plt
    cp = plt.contour(xv, yv, err, range(-20, 20, 1))
    plt.clabel(cp, inline=True, fontsize=11)
    plt.title(f'p = {p}')
    plt.xlabel('Real')
    plt.ylabel('Imag')
    ax = plt.gca()
    ax.axis('equal')
    ax.set_aspect('equal', 'box')
    plt.show()
