from firedrake import *
import firedrake.cython.dmcommon as dmcommon
from FIAT.reference_element import UFCInterval
from FIAT.quadrature import GaussLegendreQuadratureLineRule
from finat.point_set import PointSet, GaussLegendrePointSet
from finat.quadrature import QuadratureRule, TensorProductQuadratureRule
import numpy as np
import warnings
warnings.filterwarnings("ignore", category=UserWarning)
from petsc4py import PETSc
import time
import os


def make_quad_rule(N, iii):
    assert len(iii) == 3  # 3D only
    quad_rules = []
    for i in iii:
        assert i in (0, 1)
        # i == 0 -> full integration
        # i == 1 -> reduced integration
        q = N + (1 - i)
        fiat_rule = GaussLegendreQuadratureLineRule(UFCInterval(), q)
        point_set = GaussLegendrePointSet(fiat_rule.get_points())
        quad_rule = QuadratureRule(point_set, fiat_rule.get_weights())
        quad_rules.append(quad_rule)
    return TensorProductQuadratureRule(quad_rules)


def get_profile_1d(mesh, Lphys, h, L, s, gammas, d):
    V = FunctionSpace(mesh, "DQ", 0)
    f = Function(V)
    f_ = Function(V)
    coords = [0.] + [Lphys[d] + h * i for i in range(L)]
    x = SpatialCoordinate(mesh)
    for coord, gamma in zip(coords, gammas):
        f.interpolate(conditional(x[d] > coord, gamma, f_))
        f_.assign(f)
    return f


def get_profile_3d(mesh, Lphys, h, N, L, s):
    gammas = [1.]
    fiat_rule = GaussLegendreQuadratureLineRule(UFCInterval(), L)
    phis = [coord * pi / 2 for coord, in fiat_rule.get_points()]
    gammas = [1.] + [(cos(phi) * s + sin(phi)**2 / cos(phi)) / (N + 1) * h for phi in reversed(phis)]
    return tuple(get_profile_1d(mesh, Lphys, h, L, s, gammas, d) for d in range(3))


def mark_subdomain(mesh, Lphys, iii):
    subdomain_id = 8000 + 100 * iii[0] + 10 * iii[1] + 1 * iii[2]
    DQ0 = FunctionSpace(mesh, "DQ", 0)
    x = SpatialCoordinate(mesh)
    conds = []
    for i in range(3):
        conds.append(x[i] > Lphys[i] if iii[i] else x[i] < Lphys[i])
    indicator_function = Function(DQ0).interpolate(conditional(And(conds[0], And(conds[1], conds[2])), 1, 0))
    mesh.mark_entities(indicator_function, dmcommon.CELL_SETS_LABEL, subdomain_id)


load_from_file = False  # load solution from file.
mesh_name = "m"
ss = [4. * 1j,            # case 0
      1. / 4. + 4. * 1j,  # case 1
      4. + 1. / 4. * 1j   # case 2
     ]
# Direct solver:
solver_parameters = {'ksp_type': 'preonly',
                     'pc_type': 'lu',
                     'pc_factor_mat_solver_type': 'mumps',
                     'mat_mumps_icntl_22': 1}
# Iterative solver:
"""
solver_parameters = {"mat_type": "matfree",
                     "ksp_type": "gmres",
                     "pc_type": "python",
                     "pc_python_type": "firedrake.AssembledPC",
                     "assembled_pc_type": "bjacobi",
                     "assembled_pc_bjacobi_local_blocks": 1,
                     "assembled_sub_ksp_type": "preonly",
                     "assembled_sub_pc_type": "ilu",  # "ilu" or "sor"
                     "ksp_rtol": 1e-12,
                     "ksp_max_it": 1000000000,
                     "ksp_view": None,
                     "ksp_monitor": None}
"""
# Experiment case ID: [0, 1, 2]
for case in [0]:
    s = ss[case]
    PETSc.Sys.Print(f"\ncase = {case}\n")
    # FE degree: [1, 2, 3, 4]
    for N in [1]:
        PETSc.Sys.Print(f"\nN = {N}\n")
        # Refinement level: [1, 2, 3, 4]
        for nref in [1]:
            PETSc.Sys.Print(f"\nnref = {nref}\n")
            # Num. absorbing layers: range(12 // N, 0, -1)
            for L in range(12 // N, 0, -1):
                PETSc.Sys.Print(f"\nL = {L}\n")
                h5file_name = f"freq_solution_case_{case}_N_{N}_nref_{nref}_L_{L}.h5"
                txtfile_name = f"freq_error_case_{case}_N_{N}_nref_{nref}_L_{L}.txt"
                if not load_from_file:
                    PETSc.Sys.Print("Making mesh ...")
                    start = time.time()
                    Lphys = [4., 2., 2.]
                    Lhole = [1., 1., 1.]
                    h = 1. / 2**nref
                    Lcomp = [lphys + L * h for lphys in Lphys]
                    Ncomp = [round(lcomp / h) for lcomp in Lcomp]
                    mesh = BoxMesh(*Ncomp, *Lcomp, hexahedral=True, distribution_parameters={"overlap_type": (DistributedMeshOverlapType.NONE, 0)})
                    mesh.init()
                    # We assume that the reference (x, y, z) are mapped to physical (x, y, z).
                    assert all(mesh.entity_orientations[:, -1] == 0)
                    x, y, z = SpatialCoordinate(mesh)
                    DQ0 = FunctionSpace(mesh, "DQ", 0)
                    indicator_function = Function(DQ0).interpolate(conditional(x > Lhole[0], 1,
                                                                   conditional(y > Lhole[1], 1,
                                                                   conditional(z > Lhole[2], 1, 0))))
                    mesh.mark_entities(indicator_function, dmcommon.CELL_SETS_LABEL, 999)
                    mesh = Submesh(mesh, dmcommon.CELL_SETS_LABEL, 999, mesh.topological_dimension(), name=mesh_name)
                    for iii in np.ndindex(2, 2, 2):
                        mark_subdomain(mesh, Lphys, iii)
                    end = time.time()
                    PETSc.Sys.Print(f"{end - start}[s]")
                    PETSc.Sys.Print("Making profile ...")
                    start = time.time()
                    profx, profy, profz = get_profile_3d(mesh, Lphys, h, N, L, s)
                    end = time.time()
                    PETSc.Sys.Print(f"{end - start}[s]")
                    PETSc.Sys.Print("Making form ...")
                    start = time.time()
                    V = FunctionSpace(mesh, "Q", N)
                    v = TestFunction(V)
                    u = TrialFunction(V)
                    a = zero()
                    DQ0 = FunctionSpace(mesh, "DQ", 0)
                    detLam = 1. / (profx * profy * profz)
                    diag = as_tensor([[profx**2,        0,        0],
                                      [       0, profy**2,        0],
                                      [       0,        0, profz**2]])
                    for iii in np.ndindex(2, 2, 2):
                        subdomain_id = 8000 + 100 * iii[0] + 10 * iii[1] + 1 * iii[2]
                        quad_rule = make_quad_rule(N, iii)
                        a += (s**2 * inner(u, v) + inner(diag * grad(u), grad(v))) * detLam * dx(subdomain_id=subdomain_id, scheme=quad_rule)
                    RHS = Constant(0) * conj(v) * dx
                    x, y, z = SpatialCoordinate(mesh)
                    r = sqrt(x**2 + y**2 + z**2)
                    g_expr = exp(s * (1 - r)) / r
                    g = Function(V, name="target").interpolate(g_expr)
                    bc7 = DirichletBC(V, g, 7)
                    bc246 = DirichletBC(V, Constant(0.0), (2, 4, 6))
                    u = Function(V, name="solution")
                    end = time.time()
                    PETSc.Sys.Print(f"{end - start}[s]")
                    PETSc.Sys.Print("Solving ...")
                    start = time.time()
                    problem = LinearVariationalProblem(a, RHS, u, bcs=[bc7, bc246])
                    solver = LinearVariationalSolver(problem, solver_parameters=solver_parameters)
                    for _ in range(1):
                        try:
                            solver.solve()
                        except ConvergenceError:
                            pass
                        with CheckpointFile(h5file_name, "w") as afile:
                            afile.save_function(u)
                    end = time.time()
                    PETSc.Sys.Print(f"{end - start}[s]")
                    if solver.snes.getConvergedReason() <= 0:
                        PETSc.Sys.Print(f"Did not converge : intermediate solution saved in {h5file_name}")
                        continue
                else:
                    PETSc.Sys.Print(f"Loading from file ...")
                    with CheckpointFile(h5file_name, "r") as afile:
                        PETSc.Sys.Print(f"Loading mesh ...")
                        mesh = afile.load_mesh(mesh_name)
                        PETSc.Sys.Print(f"Loading function ...")
                        u = afile.load_function(mesh, "solution")
                        PETSc.Sys.Print(f"Loading complete")
                    x, y, z = SpatialCoordinate(mesh)
                    r = sqrt(x**2 + y**2 + z**2)
                    g_expr = exp(s * (1 - r)) / r
                    g = Function(u.function_space(), name="target").interpolate(g_expr)
                error = sqrt(assemble(inner(u - g_expr, u - g_expr) * dx(8000))) / sqrt(assemble(inner(g_expr, g_expr) * dx(8000)))
                error1 = sqrt(assemble(inner(grad(u - g_expr), grad(u - g_expr)) * dx(8000))) / sqrt(assemble(inner(grad(g_expr), grad(g_expr)) * dx(8000)))
                interp_error = sqrt(assemble(inner(g - g_expr, g - g_expr) * dx(8000))) / sqrt(assemble(inner(g_expr, g_expr) * dx(8000)))
                interp_error1 = sqrt(assemble(inner(grad(g - g_expr), grad(g - g_expr)) * dx(8000))) / sqrt(assemble(inner(grad(g_expr), grad(g_expr)) * dx(8000)))
                PETSc.Sys.Print(f"error        : {error}")
                PETSc.Sys.Print(f"error1       : {error1}")
                PETSc.Sys.Print(f"interp_error : {interp_error}")
                PETSc.Sys.Print(f"interp_error1: {interp_error1}")
                if mesh.comm.rank == 0:
                    with open(txtfile_name, "w") as afile:
                        afile.write(f"{error.real}")
                        assert abs(error.imag) < 1.e-15
                # pgfplot(u, "main_fig_example_3d_freq_solution.txt", print_latex_example=False)
                # pgfplot(g, "main_fig_example_3d_freq_target.txt", print_latex_example=False)
